/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.kafka.connect.mirror.utils;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Optional;
import java.util.concurrent.Callable;

public class Retrier<T> {

    private static final Logger log = LoggerFactory.getLogger(Retrier.class);
    private final Callable<T> callable;
    private final RingBuffer<RetryStatus> statusList;
    private final String label;
    private final long delayMs;

    private int maxTries = 1;
    private int attempt = 0;

    private Retrier(Callable<T> callable, RingBuffer<RetryStatus> statusList, String taskName, long delayMs) {
        this.callable = callable;
        this.statusList = statusList;
        this.label = "[" + taskName + "]";
        this.delayMs = delayMs;
    }

    public static <R> Retrier<R> of(Callable<R> callable, RingBuffer<RetryStatus> statusList, String what, long delayMs) {
        return new Retrier<>(callable, statusList, what, delayMs);
    }

    public Optional<T> retry(int maxTries) {
        this.maxTries = maxTries;
        ImmutablePair<RetryStatus, T> lastAttemptedRun;

        do {
            lastAttemptedRun = run();
            afterTry(lastAttemptedRun.getKey());
        } while (shouldTry(lastAttemptedRun.getKey()));

        if (lastAttemptedRun.getKey().isSuccess()) {
            log.info("{} is successful, result: {}", label, lastAttemptedRun.getValue());
        } else {
            log.error("{} failed with {} after {} attempt", label, lastAttemptedRun.getKey(), attempt);
        }

        return Optional.ofNullable(lastAttemptedRun.getValue());
    }

    private ImmutablePair<RetryStatus, T> run() {
        RetryStatus status;
        T result = null;
        log.info("{} trying for the {}. time", label, attempt);

        try {
            Thread.sleep(delayMs);

            result = callable.call();
            status = RetryStatus.STARTED;
            log.info("{} started successfully: ", label);
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            status = RetryStatus.INTERRUPTED;
            log.info("{} giving up, interrupted", label);
        } catch (Exception e) {
            log.error("Exception thrown in {}. try.", attempt, e);
            status = RetryStatus.FAILED;
            log.info("{} exception caught, can retry", label);
        }

        return ImmutablePair.of(status, result);
    }

    private boolean shouldTry(RetryStatus lastStatus) {
        return (maxTries == -1 || attempt < maxTries) &&
                lastStatus.isRetriable();
    }

    private void afterTry(RetryStatus lastStatus) {
        attempt += 1;
        statusList.add(lastStatus);
    }

}
